;    This file is part of CrabEmu BIOS.
;
;    Copyright (C) 2008, 2009 Lawrence Sebald
;
;    CrabEmu BIOS is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License version 2 
;    as published by the Free Software Foundation.
;
;    CrabEmu BIOS is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with CrabEmu BIOS; if not, write to the Free Software
;    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

; Some constants to start off with.
    .define         ROM_Addr    $0000
    .define         IRQ_Addr    $0038
    .define         NMI_Addr    $0066
    .define         VRAM_Len    $4000
    .define         RAM_Addr    $C000
    .define         RAM_Len     $2000
    .define         Stack_Top   $DFF0
    .define         VDP_Data    $BE
    .define         VDP_Ctl     $BF
    .define         IO_A        $DC
    .define         IO_B        $DD
    .define         Memctl      $3E
    .define         End_BIOS    $3FF
    .define         Memctl_Cart $AB

; And some variables that are useful to us.
    .enum $C000
    Header_Return   db
    Control_Data    db
    .ende

; Set up our default memory map.
    .memorymap
    defaultslot     0
    slotsize        $400
    slot            0           $0000
    .endme

    .rombankmap
    bankstotal      1
    banksize        $400
    banks           1
    .endro

; Now that all the setup work is done, here comes the code.
; The beginning of ROM. The entry point for the program.
    .bank 0 slot 0
    .org    ROM_Addr
    di
    im      1
    xor     a
    ld      (Control_Data), a
    jr      __start

; Enable the VDP, but disables vblank interrupts.
enable_vdp_noint:
    ld      a, %11000000
    out     (VDP_Ctl), a
    ld      a, $81
    out     (VDP_Ctl), a
    ret

; Enable the VDP, as well as vblank interrupts.
enable_vdp_int:
    ld      a, %11100000
    out     (VDP_Ctl), a
    ld      a, $81
    out     (VDP_Ctl), a
    ret

; Disable the VDP.
disable_vdp:
    ld      a, %10000000
    out     (VDP_Ctl), a
    ld      a, $81
    out     (VDP_Ctl), a
    ret

; IRQ Handler - Acknowledge the IRQ and grab any new data from the controller.
    .org    IRQ_Addr
irq_hnd:
    push    af
    in      a, (VDP_Ctl)        ; Figure out the type of IRQ and acknowledge it.
    bit     7, a
    jr      z, +                ; Ignore line interrupts.
    in      a, (IO_A)           ; Grab the controller data.
    cpl                         ; Invert all the bits.
    ld      (Control_Data), a
+:
    pop     af
    ei
    reti

; Initialization of the VDP.
init_vdp:
    in      a, (VDP_Ctl)        ; Read the control port to reset the flag
    ld      hl, __vdp_regs      ; Set up to copy the default registers
    ld      b, __vdp_regs_end - __vdp_regs
    ld      c, VDP_Ctl
    otir                        ; Copy the default register set.
    ld      a, $00              ; Set the first sprite's y coordinate to D0
    out     (VDP_Ctl), a        ; So as to disable the sprite display.
    ld      a, $7F
    out     (VDP_Ctl), a
    ld      a, $D0
    out     (VDP_Data), a
    ret

; NMI Handler. Simply ignore NMIs.
    .org    NMI_Addr
nmi_hnd:
    retn

; The start of REAL code. 
; Start of real code.
__start:
    ld      sp, Stack_Top       ; Initialize the SP to DFF0
    call    clear_vram          ; Clear out VRAM
    call    init_vdp            ; Initialize the VDP to a sane state
    call    init_palette        ; Initialize the VDP's palette
    call    load_font           ; Load the font to VRAM
    call    draw_bumper
    call    enable_vdp_noint
    call    find_header         ; Look for a valid TMR SEGA header.
    ld      a, (Header_Return)
    and     a
    jr      nz, +
-:
    di                          ; Disable interrupts
    call    disable_vdp         ; Disable the VDP
    call    clear_vram          ; Clear VRAM
    xor     a                   ; Clear RAM
    ld      (RAM_Addr + 1), a
    ld      hl, RAM_Addr + 1
    ld      de, RAM_Addr + 2
    ld      bc, RAM_Len - 2
    ldir
    ld      hl, __trampoline    ; Set up our trampoline
    ld      de, RAM_Addr + 1
    ld      bc, __trampoline_end - __trampoline
    ldir
    jp      RAM_Addr + 1        ; And jump off the trampoline!
+:
    call    disable_vdp
    call    clear_name_table
    call    draw_error
    call    enable_vdp_int
    ei
--:
    ld      a, (Control_Data)   ; Check if the user said start anyway.
    bit     4, a
    jr      nz, -
    halt
    jr      --

; Default VDP Register values.
__vdp_regs:
    .db     %00000100, $80
    .db     %10000000, $81
    .db     %11111111, $82
    .db     %11111111, $83
    .db     %11111111, $84
    .db     %11111111, $85
    .db     %11111111, $86
    .db     %00000000, $87
    .db     %00000000, $88
    .db     %00000000, $89
    .db     %11111111, $8A
__vdp_regs_end:

; Clear all of VRAM
clear_vram:
    ld      bc, VRAM_Len
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $40
    out     (VDP_Ctl), a
-:
    xor     a
    out     (VDP_Data), a
    dec     bc
    ld      a, b
    or      c
    jr      nz, -
    ret

; Clear the VDP Name Table.
clear_name_table:
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $78
    out     (VDP_Ctl), a
    ld      bc, $700
-:
    xor     a
    out     (VDP_Data), a
    dec     bc
    ld      a, b
    or      c
    jr      nz, -
    ret

; Set the VDP palette.
init_palette:
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $C0
    out     (VDP_Ctl), a
    ld      b, __pal_end - __pal
    ld      hl, __pal
    ld      c, VDP_Data
    otir
    ret

; Load font
load_font:
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $40
    out     (VDP_Ctl), a
    ld      hl, __font
    ld      bc, __font_end - __font
-:
    ld      a, (hl)
    out     (VDP_Data), a
    xor     a
    out     (VDP_Data), a
    out     (VDP_Data), a
    out     (VDP_Data), a
    inc     hl
    dec     bc
    ld      a, b
    or      c
    jr      nz, -
    ret

; Address to string in hl, length in b, x in c, y in d
draw_string:
    ld      a, c                ; Figure out where in VRAM to start writing.
    add     a, a                ; This isn't exactly the most pretty piece
    ld      c, a                ; of code in the world, but it will do.
    ld      a, d
    add     a, a
    add     a, a
    add     a, a
    add     a, a
    add     a, a
    add     a, a
    or      c
    out     (VDP_Ctl), a
    ld      a, d
    srl     a
    srl     a
    or      $78
    out     (VDP_Ctl), a
-:
    ld      a, (hl)
    inc     hl
    sub     $2F
    out     (VDP_Data), a
    xor     a
    out     (VDP_Data), a
    djnz    -
    ret

draw_bumper:
    ld      hl, __crabemu_bios
    ld      b, 12
    ld      c, 10
    ld      d, 17
    call    draw_string
    ld      hl, __copyright_notice1
    ld      b, 12
    ld      c, 10
    ld      d, 18
    call    draw_string
    ld      hl, __copyright_notice2
    ld      b, 15
    ld      c, 8
    ld      d, 19
    call    draw_string
+:
    ret

draw_error:
    ld      hl, __error_msg1
    ld      b, 14
    ld      c, 9
    ld      d, 4
    call    draw_string
    ld      hl, __error_msg2
    ld      b, 10
    ld      c, 11
    ld      d, 10
    call    draw_string
    ld      hl, __error_msg3
    ld      b, 14
    ld      c, 9
    ld      d, 11
    call    draw_string
    call    draw_bumper
    ret

; If Z is set when this function returns, the "TMR SEGA" string was found
; at the address of hl.
check_hl:
    push    hl
    push    bc
    push    ix
    ld      bc, 8
    ld      ix, _tmr_sega_const
    ld      a, (ix + 0)         ; Grab first character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 1)         ; Grab second character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 2)         ; Grab third character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 3)         ; Grab fourth character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 4)         ; Grab fifth character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 5)         ; Grab sixth character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 6)         ; Grab seventh character from reference
    cpi
    jr      nz, +
    ld      a, (ix + 7)         ; Grab eighth character from reference
    cpi
+:
    pop     ix
    pop     bc
    pop     hl
    ret

; Attempt to find the TMR SEGA string that starts the standard SMS header.
find_header:
    ld      hl, $1FF0           ; Check at 0x1FF0 first.
    call    check_hl
    jr      z, +
    ld      hl, $3FF0           ; Then try 0x3FF0.
    call    check_hl
    jr      z, +
    ld      hl, $7FF0           ; Finally, try 0x7FF0.
    call    check_hl
    jr      z, +
    ld      a, -1               ; We haven't found a header if we get here.
    ld      (Header_Return), a  ; Set a failure flag at 0xC001
    ret
+:
    xor     a                   ; We have found a header if we get here.
    ld      (Header_Return), a  ; Set a success flag at 0xC001
    ret

; Trampoline for starting a game. This sets the memctl register to the value in
; a, and jumps unconditionally to the start of ROM. This is meant to be run from
; RAM.
__trampoline:
    ld      a, Memctl_Cart
    out     (Memctl), a
    ld      a, $00              ; Reset registers:  a - 0x00
    add     a, $00              ;                   f - 0x40
    ld      bc, $0000           ;                   bc - 0x0000
    ld      de, $0000           ;                   de - 0x0000
    ld      hl, $0000           ;                   hl - 0x0000
    jp      ROM_Addr
__trampoline_end:

; The font used in the display.
__font:
    .include "font.i"
__font_end:

; The palette used for the font.
__pal:
    .db     $00, $00, $FF, $0F
__pal_end:

; Various strings displayed by the BIOS.
__crabemu_bios:
    .db     "CRABEMU/BIOS"          ; 12

__copyright_notice1:
    .db     "[/2008:/2009"          ; 12

__copyright_notice2:
    .db     "LAWRENCE/SEBALD"       ; 15

__error_msg1:
    .db     "SOFTWARE/ERROR"        ; 14

__error_msg2:
    .db     "PRESS/1/TO"            ; 10

__error_msg3:
    .db     "START/THE/GAME"        ; 14

_tmr_sega_const:
    .db    "TMR SEGA"

; Make sure we're exactly 1024 bytes long.
    .org    End_BIOS
    .db     0
