;    This file is part of CrabEmu BIOS.
;
;    Copyright (C) 2008, 2009, 2012 Lawrence Sebald
;
;    CrabEmu BIOS is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License version 2 
;    as published by the Free Software Foundation.
;
;    CrabEmu BIOS is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with CrabEmu BIOS; if not, write to the Free Software
;    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

; Some constants to start off with.
    .define         ROM_Addr    $0000
    .define         IRQ_Addr    $0038
    .define         NMI_Addr    $0066
    .define         VRAM_Len    $4000
    .define         RAM_Addr    $C000
    .define         RAM_Len     $2000
    .define         Stack_Top   $DFF0
    .define         VDP_Data    $BE
    .define         VDP_Ctl     $BF
    .define         IO_A        $DC
    .define         IO_B        $DD
    .define         Memctl      $3E
    .define         IOctl       $3F
    .define         SDSCData    $FD
    .define         PSG         $7F
    .define         CHKSUM_Addr RAM_Addr + (RAM_Len / 2)

; And some variables that are useful to us.
    .enum $C000
    Last_Memctl     db
    Checksum_Return db
    Control_Data    db
    Region_Code     db
    .ende

; Set up our default memory map.
    .memorymap
    defaultslot     0
    slotsize        $4000
    slot            0           $0000
    slot            1           $4000
    .endme

    .rombankmap
    bankstotal      2
    banksize        $4000
    banks           2
    .endro

; Add a SDSC (and standard SMS) tag.
    .sdsctag        1.0, "CrabEmu BIOS", "SMS BIOS Workalike", "Lawrence Sebald"

; Now that all the setup work is done, here comes the code.
; The beginning of ROM. The entry point for the program.
    .bank 0 slot 0
    .org    ROM_Addr
    di
    im      1
    xor     a
    ld      (Control_Data), a
    jr      __start

; Enable the VDP, but disables vblank interrupts.
enable_vdp_noint:
    ld      a, %11000000
    out     (VDP_Ctl), a
    ld      a, $81
    out     (VDP_Ctl), a
    ret

; Enable the VDP, as well as vblank interrupts.
enable_vdp_int:
    ld      a, %11100000
    out     (VDP_Ctl), a
    ld      a, $81
    out     (VDP_Ctl), a
    ret

; Disable the VDP.
disable_vdp:
    ld      a, %10000000
    out     (VDP_Ctl), a
    ld      a, $81
    out     (VDP_Ctl), a
    ret

; IRQ Handler - Acknowledge the IRQ and grab any new data from the controller.
    .org    IRQ_Addr
irq_hnd:
    push    af
    in      a, (VDP_Ctl)        ; Figure out the type of IRQ and acknowledge it.
    bit     7, a
    jr      z, +                ; Ignore line interrupts.
    in      a, (IO_A)           ; Grab the controller data.
    cpl                         ; Invert all the bits.
    ld      (Control_Data), a
+:
    pop     af
    ei
    reti

; Initialization of the VDP.
init_vdp:
    in      a, (VDP_Ctl)        ; Read the control port to reset the flag
    ld      hl, __vdp_regs      ; Set up to copy the default registers
    ld      b, __vdp_regs_end - __vdp_regs
    ld      c, VDP_Ctl
    otir                        ; Copy the default register set.
    ld      a, $00              ; Set the first sprite's y coordinate to D0
    out     (VDP_Ctl), a        ; So as to disable the sprite display.
    ld      a, $7F
    out     (VDP_Ctl), a
    ld      a, $D0
    out     (VDP_Data), a
    ret

; NMI Handler. Simply ignore NMIs.
    .org    NMI_Addr
nmi_hnd:
    retn

; The start of REAL code. 
__start:
    ld      sp, Stack_Top       ; Initialize the SP to DFF0
    call    init_psg            ; Stop any sound playing previously.
    call    clear_vram          ; Clear VRAM
    call    check_region        ; Determine the region we're running on
    ld      (Region_Code), a    ; Save the region to RAM
    call    write_sdsc          ; Write a test message to SDSC terminal.
    call    init_vdp            ; Initialize the VDP to a sane state
    call    init_palette        ; Initialize the VDP's palette
    call    load_font           ; Load the font to VRAM
    call    draw_bumper
    call    enable_vdp_noint
    call    copy_checksum       ; Copy the checksumming code to RAM
    call    CHKSUM_Addr         ; Call the checksumming code
    ld      a, (Checksum_Return)
    and     a
    jr      nz, +
-:
    di                          ; Disable interrupts
    call    disable_vdp         ; Disable the VDP
    call    clear_vram          ; Clear VRAM
    ld      a, $F0              ; Clear RAM to 0xF0
    ld      (RAM_Addr + 1), a
    ld      hl, RAM_Addr + 1
    ld      de, RAM_Addr + 2
    ld      bc, RAM_Len - 2
    ldir
    ld      a, (Last_Memctl)    ; Set up our trampoline
    ld      hl, __trampoline
    ld      de, RAM_Addr + 1
    ld      bc, __trampoline_end - __trampoline
    ldir
    jp      RAM_Addr + 1        ; And jump off the trampoline!
+:
    call    disable_vdp
    call    clear_name_table
    call    draw_error
    call    enable_vdp_int
    ei
--:
    ld      a, (Control_Data)   ; Check if the user said start anyway.
    bit     4, a
    jr      nz, -
    halt
    jr      --

; Init the PSG
init_psg:
    ld      hl, __psg_regs      ; Set up to copy the default registers
    ld      b, __psg_regs_end - __psg_regs
    ld      c, PSG
    otir                        ; Copy the default register set.
    ret

; Default PSG Register values.
__psg_regs:
    .db     %10011111               ; Silence channel 0
    .db     %10111111               ; Silence channel 1
    .db     %11011111               ; Silence channel 2
    .db     %11111111               ; Silence noise
    .db     %10000000, %00000000    ; Tone 0 = 0
    .db     %10100000, %00000000    ; Tone 1 = 0
    .db     %11000000, %00000000    ; Tone 2 = 0
    .db     %11100011               ; Noise shift rate = Tone 2 = 0
__psg_regs_end:

; Default VDP Register values.
__vdp_regs:
    .db     %00000100, $80
    .db     %10000000, $81
    .db     %11111111, $82
    .db     %11111111, $83
    .db     %11111111, $84
    .db     %11111111, $85
    .db     %11111111, $86
    .db     %00000000, $87
    .db     %00000000, $88
    .db     %00000000, $89
    .db     %11111111, $8A
__vdp_regs_end:

; Clear all of VRAM
clear_vram:
    ld      bc, VRAM_Len
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $40
    out     (VDP_Ctl), a
-:
    xor     a
    out     (VDP_Data), a
    dec     bc
    ld      a, b
    or      c
    jr      nz, -
    ret

; Clear the VDP Name Table.
clear_name_table:
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $78
    out     (VDP_Ctl), a
    ld      bc, $700
-:
    xor     a
    out     (VDP_Data), a
    dec     bc
    ld      a, b
    or      c
    jr      nz, -
    ret

; Check the region of the SMS. Returns 0 for Domestic, 1 for export.
check_region:
    ld      a, %11110101        ; Configure both TH pins as output
    out     (IOctl), a          ; and output 1s on both pins
    in      a, (IO_B)           ; Read the IO Port B/misc
    and     %11000000           ; Mask out anything except the TH values
    cp      %11000000           ; Check if both are set
    jr      nz, +               ; If not, we're on a Japanese system.
    ld      a, %01010101        ; Configure both TH pins as output
    out     (IOctl), a          ; and output 0s on both pins
    in      a, (IO_B)           ; Read the IO Port B/misc
    and     %11000000           ; Mask out anything other than the TH pins
    jr      nz, +               ; If either are set, this is a Japanse system.
    ld      a, %11111111        ; Reset everything to the defaults
    out     (IOctl), a
    ld      a, 1
    ret
+:
    xor     a
    ret
    

; Copy checksumming code to RAM
copy_checksum:
    ld      a, (Region_Code)
    and     a
    jr      z, +
    ld      hl, __checksum_bin
    ld      de, CHKSUM_Addr
    ld      bc, checksum_size
    ldir
    ret
+:
    ld      hl, __domestic_bin
    ld      de, CHKSUM_Addr
    ld      bc, domestic_size
    ldir
    ret

; Set the VDP palette.
init_palette:
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $C0
    out     (VDP_Ctl), a
    ld      b, __pal_end - __pal
    ld      hl, __pal
    ld      c, VDP_Data
    otir
    ret

; Load font
load_font:
    ld      a, $00
    out     (VDP_Ctl), a
    ld      a, $40
    out     (VDP_Ctl), a
    ld      hl, __font
    ld      bc, __font_end - __font
-:
    ld      a, (hl)
    out     (VDP_Data), a
    xor     a
    out     (VDP_Data), a
    out     (VDP_Data), a
    out     (VDP_Data), a
    inc     hl
    dec     bc
    ld      a, b
    or      c
    jr      nz, -
    ret

; Address to string in hl, length in b, x in c, y in d
draw_string:
    ld      a, c                ; Figure out where in VRAM to start writing.
    add     a, a                ; This isn't exactly the most pretty piece
    ld      c, a                ; of code in the world, but it will do.
    ld      a, d
    add     a, a
    add     a, a
    add     a, a
    add     a, a
    add     a, a
    add     a, a
    or      c
    out     (VDP_Ctl), a
    ld      a, d
    srl     a
    srl     a
    or      $78
    out     (VDP_Ctl), a
-:
    ld      a, (hl)
    inc     hl
    sub     $20
    out     (VDP_Data), a
    xor     a
    out     (VDP_Data), a
    djnz    -
    ret

draw_bumper:
    ld      hl, __crabemu_bios
    ld      b, 12
    ld      c, 10
    ld      d, 21
    call    draw_string
    ld      hl, __copyright_notice
    ld      b, 28
    ld      c, 2
    ld      d, 22
    call    draw_string
    ld      a, (Region_Code)
    and     $FF
    jr      nz, +
    ld      hl, __region_japanese
    ld      b, 13
    ld      c, 9
    ld      d, 23
    call    draw_string
+:
    ret

draw_error:
    ld      hl, __error_msg1
    ld      b, 14
    ld      c, 9
    ld      d, 4
    call    draw_string
    ld      hl, __error_msg4
    ld      b, 26
    ld      c, 3
    ld      d, 10
    call    draw_string
    call    draw_bumper
    ld      a, (Checksum_Return)
    ld      d, 6
    cp      -1
    jr      z, +
    cp      -3
    jr      z, ++
    ld      hl, __error_msg3
    ld      b, 16
    ld      c, 8
    call    draw_string
    ret
+:
    ld      hl, __error_msg2
    ld      b, 19
    ld      c, 6
    call    draw_string
    ret
++:
    ld      hl, __error_msg5
    ld      b, 13
    ld      c, 9
    call    draw_string
    ret

write_sdsc:
    ld      a, $e7
    out     (Memctl), a
    ld      a, 'C'
    out     (SDSCData), a
    ld      a, 'r'
    out     (SDSCData), a
    ld      a, 'a'
    out     (SDSCData), a
    ld      a, 'b'
    out     (SDSCData), a
    ld      a, 'E'
    out     (SDSCData), a
    ld      a, 'm'
    out     (SDSCData), a
    ld      a, 'u'
    out     (SDSCData), a
    ld      a, ' '
    out     (SDSCData), a
    ld      a, 'B'
    out     (SDSCData), a
    ld      a, 'I'
    out     (SDSCData), a
    ld      a, 'O'
    out     (SDSCData), a
    ld      a, 'S'
    out     (SDSCData), a
    ld      a, $e3
    out     (Memctl), a
    ret
    

; Trampoline for starting a game. This sets the memctl register to the value in
; a, and jumps unconditionally to the start of ROM. This is meant to be run from
; RAM.
__trampoline:
    out     (Memctl), a
    ld      a, $00              ; Reset registers:  a - 0x00
    add     a, $00              ;                   f - 0x40
    ld      bc, $0000           ;                   bc - 0x0000
    ld      de, $0000           ;                   de - 0x0000
    ld      hl, $0000           ;                   hl - 0x0000
    jp      ROM_Addr
__trampoline_end:

; The export SMS checksumming code. This code is meant to be run from RAM. Its
; origin address is halfway through RAM (D000).
__checksum_bin:
    .incbin "checksum.bin" fsize checksum_size

; The domestic SMS slot checking code. This code is meant to be run from RAM.
; Its origin address is halfway through RAM (D000).
__domestic_bin:
    .incbin "domesticcheck.bin" fsize domestic_size

; The font used in the display.
__font:
    .include "font.i"
__font_end:

; The palette used for the font.
__pal:
    .db     $00, $3F, $00, $00, $00, $00, $00, $00
    .db     $00, $00, $00, $00, $00, $00, $00, $00
    .db     $00, $00, $00, $00, $00, $00, $00, $00
    .db     $00, $00, $00, $00, $00, $00, $00, $00
__pal_end:

; Various strings displayed by the BIOS.
__crabemu_bios:
    .db     "CrabEmu BIOS"

__copyright_notice:
    .db     $7F
    .db     " 2008, 2009 Lawrence Sebald"

__region_japanese:
    .db     "Domestic Mode"         ; 13

__error_msg1:
    .db     "Software Error"        ; 14

__error_msg2:
    .db     "Missing SEGA Header"   ; 19

__error_msg3:
    .db     "Invalid Checksum"      ; 16

__error_msg4:
    .db     "Press 1 to start the game."    ; 26

__error_msg5:
    .db     "No data found"         ; 13
